// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
/*
 attmove

   Authors:  John Hoar - ESA SCI-SDX
         
   Description: Create new Attitude History File to simulate tracking of a moving target
 
   Comments: 

   Version 1.0 - Version uploaded
 
  $Id: attmove.cc,v 1.3 2004/03/05 09:31:13 jhoar Exp $
*/

#include "attmove.h"

#include <Tracer.h>
#include <errstr.h>
#include "SmartDal"
#include "param.h"
#include "STime.h"
#include "slatec++.h"
#include "CalOalUtility.h"
#include "PhysicalConstants.h"
#include "OalServer.h"
#include "Celestial.h"
#include "Xmm.h"

// Class declarations
class DiagnosticFile {

public:

  DiagnosticFile(std::string filename);
  ~DiagnosticFile();
 
  void createDiagnosticFile(unsigned nRows);
  
  void writeDiagnostics(unsigned row,
			STime time,
			EquatorialDirection &point,
			EquatorialDirection &object,
			EquatorialDirection &correct,
			EquatorialDirection &final);
  
  void closeDiagnostics();

private:
  
  DataSet *ds;
  Table *tab;

  bool isInit;

  real64 *timed;
  real64 *xracd;
  real64 *xdecd;
  real64 *oracd;
  real64 *odecd;
  real64 *pracd;
  real64 *pdecd;
  real64 *fracd;
  real64 *fdecd;
  real64 *cracd;
  real64 *cdecd;

};


// Internal function prototypes
void expand(std::string inputAHF,std::string outputAHF, real32 minStable, real32 granularity);
void correct(std::string outputAHF, real32 minStable, real32 granularity); 
void process(std::string outputAHF, std::string inputEPH, EquatorialDirection refPoint, DiagnosticFile *diag);
		      
void attmove ()
{
  PROC

  // Read Parameters
  // inputAHF - AHF to process 
  std::string inputAHF = stringParameter("input");
  
  // inputEPH - Ephemris file to process 
  std::string inputEPH = stringParameter("track");
  
  // outputAHF - AHF to generate 
  std::string outputAHF = stringParameter("output");

  // minStable - minimum duration of stable pointing to extract 
  real32 minStable = realParameter("minstable");

  // granularity - duration of generated stable pointing bins 
  real32 granularity = realParameter("granularity");

  // refra - right ascension of reference point
  real32 refra;
  // refdec - declination of reference point
  real32 refdec;

  // withref - is a reference point specified?
  bool withref = booleanParameter("withrefatt");

  // Read ref parameters
  if ( withref ) {
    refra = realParameter("refra");
    refdec = realParameter("refdec");
  }
  else {
    ProposalInfo pi = oalServer.odf()->proposalInfo();
    refra = pi.targetRa;
    refdec = pi.targetDec;
  }

  bool diag = booleanParameter("creatediagnostics");
  DiagnosticFile *diagFile = 0;	

  if ( diag ) { 
  	diagFile = new DiagnosticFile(stringParameter("diagfile"));
  }

  // Create reference point object
  EquatorialDirection refPoint(RightAscension(refra * DEG_TO_RAD), 
			       Declination(refdec * DEG_TO_RAD));

  // Check that the size of the output units is smaller that then 
  // periods to be processed
  if (granularity > minStable) {
    Sas::error("invalidArgs","Granularity must be smaller than stable periods");
  }

  // Execute the individual steps
  expand(inputAHF, outputAHF, minStable, granularity);
  correct(outputAHF, minStable, granularity);
  process(outputAHF, inputEPH, refPoint, diagFile);

  ENDPROC

}  

// Creates a new file as a clone of the original
// and creates new rows where the duration of the stable 
// pointing period SPP is large enough to be broken down into
// smaller entries
void expand(std::string inputAHF,std::string outputAHF, real32 minStable, real32 granularity) {  

  unsigned int rows2add = 0;
  unsigned int sourRow = 0;
  unsigned int destRow = 1;
  real32 pPeriod = 0.0;
  real32 reps = 0.0;

  // Duplicate $outputAHF from $inputAHF read/write
  dataSetServer->copy(inputAHF,outputAHF);

  // SmartDal is useful
  using namespace SmartDal;

  // Open outputAHF RO
  DataSetH inputAHFds = SmartDal::open_set(inputAHF, Read);
  TableH inputTable = inputAHFds->table(0);
  
  // Open outputAHF RW
  DataSetH outputAHFds = SmartDal::open_set(outputAHF, Modify);
  TableH outputTable = outputAHFds->table(0);
  
  // Set up table access
  real32 *iVALDUR = inputTable->column("VALDUR")->data()->real32Data();
  
  unsigned int nRows = outputTable->rows();
  for (unsigned int n = 0; n < nRows; n++, sourRow++, destRow++) {

    // Get stable pointing period (SPP)
    pPeriod = iVALDUR[n];

    // If this SPP is long enough to be processed
    if (pPeriod >= minStable) {

      // Calculate rows to add
      reps = pPeriod / granularity;
      rows2add = (int)reps;

      // If SPP is exactly divisible by the size, 1 less row to add
      if ((reps - rows2add) == 0.0) {
	rows2add--;
      }
      
      // Add the rows to the table
      // Process rows 
      if (rows2add > 0) {

	// Insert the rows
	outputTable->insertRows(destRow,rows2add);

	// While we still have data to process
	for (unsigned int m = 0; m < rows2add; m++) {

	  // Copy values of the original row to the new rows
	  outputTable->copyRows(sourRow,destRow + m,1);
	  
	} // for

	// Increment counters
	sourRow += rows2add;
	destRow += rows2add;

      } // if

    } // if

  } // for

} // expand

// The VALTIME and VALDUR entries in the AHF need to be updated
// where they are derived from a original large entry
void correct(std::string outputAHF, real32 minStable, real32 granularity) {  

  real32 pPeriod = 0.0;
  real32 reps = 0.0;
  real32 durationProcessed = 0.0;
  unsigned int rows2process = 0;

  // SmartDal is useful
  using namespace SmartDal;

  // Open outputAHF RW
  DataSetH outputAHFds = SmartDal::open_set(outputAHF, Modify);
  TableH outputTable = outputAHFds->table(0);
  
  // Set up table access
  real32 *iVALDUR = outputTable->column("VALDUR")->data()->real32Data();  
  ColumnH outputVALTIMEc = outputTable->column("VALTIME");

  unsigned int nRows = outputTable->rows();
  for (unsigned int n = 0; n < nRows;) {

    // Initialise vars
    rows2process = 0;

    // Get the time and the duration
    pPeriod = iVALDUR[n];
    STime pTime(SmartDal::get_string_row(outputVALTIMEc,n));
    
    // If this pointing period is long enough to be processed
    if (pPeriod >= minStable) {

      // Reset time processed
      durationProcessed = 0.0;

      // Calculate number of rows to be processed
      reps = pPeriod / granularity;
      rows2process = (int)reps;

      // If SPP is not exactly divisible by the size, 1 more row to add      
      if ((reps - rows2process) != 0.0) {
	rows2process++;
      }
      
      // Add the rows to the table
      // While we still have data to process
      if (rows2process > 0) {

	// For each row to process
	for (unsigned int m = 0; m < rows2process; m++) {

	  // Update the duration
	  if (durationProcessed + granularity < pPeriod) {
	    iVALDUR[n + m] = granularity;
	  }
	  else {
	    // Last row, so duration is remaining time
	    iVALDUR[n + m] = pPeriod - durationProcessed;
	  }

	  // Update the time
	  std::string time = (pTime + (m * granularity)).toString(0) + 'Z';
	  SmartDal::set_string_row(outputVALTIMEc, n + m, time);

	  // Increase time processed
	  durationProcessed += granularity;

	} // for

      } // if

    } // if

    // Depending on what we have done, increase the row counter
    if (rows2process > 0) {
      n += rows2process;
    }
    else {
      n++;
    }
    
  } // for

} // correct()

void process(std::string outputAHF, std::string inputEPH, EquatorialDirection refPoint, DiagnosticFile *diag) {

  // Open Object track file
  DataSet *objTrackSet = dataSetServer->open(inputEPH, Dal::Read);
  Table *trackTab = objTrackSet->table(0);

  // Read object track from data file
  std::vector<double> mjdvec, ravec, decvec, deltavec;
  CalOalUtility::fillVector(trackTab, "MJD", mjdvec);
  CalOalUtility::fillVector(trackTab, "RA", ravec);
  CalOalUtility::fillVector(trackTab, "DEC", decvec);
  CalOalUtility::fillVector(trackTab, "DELTA", deltavec);	//	[AU]

  // Close data file
  dataSetServer->close(objTrackSet);
  
  // Set up interpolator for vectors
  Slatec::CubicHermiteInterpolant<double> 
    objra(mjdvec, ravec),
    objdec(mjdvec, decvec),
    delta(mjdvec, deltavec);

  // Obtain PositionLocator
  const PositionLocator *xmmPosLoc = oalServer.positionLocator();

  // SmartDal is useful
  using namespace SmartDal;  

  // Open outputAHF RW
  DataSetH outputAHFds = SmartDal::open_set(outputAHF, Modify);
  TableH outputTable = outputAHFds->table(0);

  // Set up table access
  ColumnH outputVALTIMEc = outputTable->column("VALTIME");
  ColumnH outputVIEWRAc = outputTable->column("VIEWRA");
  ColumnH outputVIEWDECLc = outputTable->column("VIEWDECL");

  unsigned nRows = outputTable->rows();

  if( diag ) {
    diag->createDiagnosticFile(nRows);
  }

  // Process AHF entries
  for (unsigned int n = 0; n < nRows; n++) {
    
    // Get the time of the entry
    STime rowTime(SmartDal::get_string_row(outputVALTIMEc,n));
    
    // Create the AHF Attitude
    RightAscension viewRA(SmartDal::get_string_row(outputVIEWRAc,n));
    Declination viewDEC(SmartDal::get_string_row(outputVIEWDECLc,n));
    EquatorialDirection viewPoint(viewRA, viewDEC);

    // Get the MJD time 
    double mjd = rowTime.modifiedJulianDay().full();

    // Get the Object Coordinates, interpolating from Object Track
    RightAscension objectRA(objra(mjd) * DEG_TO_RAD);
    Declination objectDEC(objdec(mjd) * DEG_TO_RAD);
	
    EquatorialDirection objectPoint(objectRA, objectDEC);
    EquatorialDirection correctedPoint(objectRA, objectDEC);
    
    // Correct Object Coordinates for geocentric parallax shift
    // Object position in km, not m
    correctedPoint.toApparentDirection(xmmPosLoc->stateVector(rowTime).position,
				       delta(mjd) * PhysicalConstants::AU / 1000.0);

    // Calculate RA and DEC of final pointing
    RightAscension finalRA(refPoint.ra().angle() + (viewPoint.ra().angle() - correctedPoint.ra().angle()));
    Declination finalDEC(refPoint.dec().angle() + (viewPoint.dec().angle() - correctedPoint.dec().angle()));

    EquatorialDirection finalPoint(finalRA, finalDEC);

    // Write RA and DEC to AHF
    SmartDal::set_string_row(outputVIEWRAc,n,finalRA.toString(':'));
    SmartDal::set_string_row(outputVIEWDECLc,n,finalDEC.toString(':'));
    
    if ( diag ) {
      diag->writeDiagnostics(n, rowTime, viewPoint, objectPoint, correctedPoint, finalPoint);
    }

  } // for
  
  if ( diag ) {
    diag->closeDiagnostics();
  }


} // process()


DiagnosticFile::DiagnosticFile(std::string filename) {

  isInit = false;
  ds = dataSetServer->open(filename, Dal::Create);
  
}

void DiagnosticFile::createDiagnosticFile(unsigned nRows) {
  
  tab = ds->addTable("ATTDIAG", nRows);
  
  timed = tab->addColumn("TIME",Column::Real64,"MJD")->data()->real64Data();
  xracd = tab->addColumn("PNTRA",Column::Real64)->data()->real64Data();
  xdecd = tab->addColumn("PNTDEC",Column::Real64)->data()->real64Data();
  oracd = tab->addColumn("OBJRA",Column::Real64)->data()->real64Data();
  odecd = tab->addColumn("OBJDEC",Column::Real64)->data()->real64Data();
  pracd = tab->addColumn("PARRA",Column::Real64)->data()->real64Data();
  pdecd = tab->addColumn("PARDEC",Column::Real64)->data()->real64Data();
  fracd = tab->addColumn("FINRA",Column::Real64)->data()->real64Data();
  fdecd = tab->addColumn("FINDEC",Column::Real64)->data()->real64Data();
  cracd = tab->addColumn("CORRA",Column::Real64, "arcsecs")->data()->real64Data();
  cdecd = tab->addColumn("CORDEC",Column::Real64, "arcsecs")->data()->real64Data();

  isInit = true;
 
}

void DiagnosticFile::writeDiagnostics(unsigned n,
				      STime time,
				      EquatorialDirection &point,
				      EquatorialDirection &object,
				      EquatorialDirection &correct,
				      EquatorialDirection &final) {
  
  
  if (isInit) {		
    timed[n] = time.modifiedJulianDay().full();
    xracd[n] = point.ra().angle();
    xdecd[n] = point.dec().angle();
    oracd[n] = object.ra().angle();
    odecd[n] = object.dec().angle();
    pracd[n] = correct.ra().angle();
    pdecd[n] = correct.dec().angle();
    fracd[n] = final.ra().angle();
    fdecd[n] = final.dec().angle();
    
    cracd[n] = (object.ra().angle() - correct.ra().angle()) * 206265;
    cdecd[n] = (object.dec().angle() - correct.dec().angle()) * 206265;

  }
  
}

void DiagnosticFile::closeDiagnostics() {
  
  dataSetServer->close(ds);
  
}

DiagnosticFile::~DiagnosticFile() {

  if (isInit) {
    delete(timed);
    delete(xracd);
    delete(xdecd);
    delete(oracd);
    delete(odecd);
    delete(pracd);
    delete(pdecd);
    delete(fracd);
    delete(fdecd);
    delete(cracd);
    delete(cdecd);
  }

}
